
// SkyCAPTCHA
// Copyright (C) 2010 S.Fuchita (soramimi)


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>
#include <vector>
#include <map>
#include <string>
#include <sstream>
#include "captcha.h"
#include "rand64.h"

static const int IDLEN = 20;

static const int TIMEOUT = 3600;

#define CAPTCHA_STATUS_FILE "captcha.status.cgi"


void parse_query_string(char const *ptr, std::map<std::string, std::string> *map)
{
	char const *left = ptr;
	char const *eq = 0;
	while (1) {
		int c = *ptr;
		if (c == 0 || c == '&') {
			if (eq) {
				if (left < eq) {
					std::string key(left, eq);
					std::string val(eq + 1, ptr);
					(*map)[key] = val;
				}
			}
			if (c == 0) {
				break;
			}
			ptr++;
			left = ptr;
			ptr = 0;
			eq = 0;
		} else if (c == '=') {
			eq = ptr;
			ptr++;
		} else {
			ptr++;
		}
	}
}

bool readfile(int fd, std::vector<std::string> *lines)
{
	struct stat st;
	if (fstat(fd, &st) != 0) {
		return false;
	}
	std::vector<char> buffer;
	buffer.resize(st.st_size);
	if (read(fd, &buffer[0], st.st_size) != st.st_size) {
		return false;
	}
	char const *ptr = &buffer[0];
	char const *end = ptr + st.st_size;
	char const *left;
	left = ptr;
	while (1) {
		int c = -1;
		if (ptr < end) {
			c = (unsigned char)*ptr;
		}
		if (c == -1 || c == '\r' || c == '\n') {
			if (left < ptr) {
				std::string line(left, ptr);
				lines->push_back(line);
			}
			if (c == -1) {
				break;
			}
			if (c == '\r') {
				ptr++;
				if (ptr < end && *ptr == '\n') {
					ptr++;
				}
			} else {
				ptr++;
			}
			left = ptr;
		} else {
			ptr++;
		}
	}
	return true;
}

static bool captcha_process(std::string const &remove_id, std::string *new_id, int generate_len)
{
	int fd = open(CAPTCHA_STATUS_FILE, O_RDWR | O_CREAT, 0644);
	if (fd == -1) {
		return false;
	}
	if (flock(fd, LOCK_EX) != 0) {
		return false;
	}
	std::vector<std::string> lines;
	if (!readfile(fd, &lines)) {
		return false;
	}

	time_t now = time(0);

	if (generate_len > 0) {

		char tmp[100];

		sprintf(tmp, "%u ", (int)now);
		int pos = strlen(tmp);
		char *id = tmp + pos;

		bool ng = false;
		do {
			for (int i = 0; i < IDLEN; i++) {
				id[i] = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"[random64() % 62];
			}
			id[IDLEN] = 0;
			for (std::vector<std::string>::const_iterator it = lines.begin(); it != lines.end(); it++) {
				if (strncmp(it->c_str(), id, IDLEN) == 0) {
					ng = true;
					break;
				}
			}
		} while (ng);

		*new_id = id;

		pos += IDLEN;
		tmp[pos] = ' ';
		pos++;

		for (int i = 0; i < generate_len; i++) {
			tmp[pos] = '0' + random64() % 10;
			pos++;
		}
		tmp[pos] = 0;

		lines.push_back(tmp);

	}

	std::stringstream ss;
	for (std::vector<std::string>::const_iterator it = lines.begin(); it != lines.end(); it++) {
		if (atoi(it->c_str()) + TIMEOUT < now) {
			goto skip;
		}
		if (!remove_id.empty()) {
			char const *p = strchr(it->c_str(), ' ');
			if (p && strncmp(p + 1, remove_id.c_str(), remove_id.size()) == 0) {
				goto skip;
			}
		}
		ss.write(it->c_str(), it->size());
		ss.write("\n", 1);
skip:;
	}

	std::string line = ss.str();
	lseek(fd, 0, SEEK_SET);
	ftruncate(fd, 0);

	write(fd, line.c_str(), line.size());
	close(fd);

	return true;
}

bool captcha_generate(int generate_len)
{
	std::string remove_id;

	std::string new_id;
	bool f = captcha_process(std::string(), &new_id, generate_len);
	if (!f) {
		return false;
	}

	puts(new_id.c_str());
	return true;
}

bool captcha_remove(std::string const &remove_id)
{
	return captcha_process(remove_id, 0, 0);
}

std::string captcha_find(char const *id)
{
	int fd = open(CAPTCHA_STATUS_FILE, O_RDWR | O_CREAT, 0644);
	if (fd == -1) {
		return std::string();
	}
	if (flock(fd, LOCK_EX) == 0) {
		std::vector<std::string> lines;
		if (readfile(fd, &lines)) {
			for (std::vector<std::string>::const_iterator it = lines.begin(); it != lines.end(); it++) {
				char const *a = strchr(it->c_str(), ' ');
				if (a) {
					char const *b = strchr(a + 1, ' ');
					if (b) {
						if (strncmp(a + 1, id, IDLEN) == 0) {
							return b + 1;
						}
					}
				}
			}
		}
	}
	close(fd);
	return std::string();
}

bool response_image(std::string const &text)
{
	int w, h;

	w = 0;
	h = 0;

	soramimi::image::Image image;
	soramimi::image::Image digits;
	soramimi::image::read_png(&digits, "captcha.digits.cgi", &w, &h);

	if (w < 1 || h < 1) {
		printf("Content-type: text/plain\r\n\r\n");
		printf("unexpected error: could not load digits image file.");
		return false;
	}

	w /= 16;
	h /= 32;
	if (w != 62) {
		return false;
	}
	if (h < 1) {
		return false;
	}
	int f = random64() % h;

	soramimi::captcha(text.c_str(), text.size(), &digits, f, &image, &w, &h);

	std::vector<unsigned char> v;
	v.reserve(65536);
	if (false) {
		soramimi::image::write_png(&v, &image, w, h);
		printf("Content-type: image/png\r\n\r\n");
	} else {
		soramimi::image::write_jpeg(&v, &image, w, h, 30);
		printf("Content-type: image/jpg\r\n\r\n");
	}
	fwrite(&v[0], 1, v.size(), stdout);
	return true;
}

int main(int argc, char const **argv)
{
	int argi;
	for (argi = 1; argi < argc; argi++) {
		char const *arg = argv[argi];
		if (strcmp(arg, "generate") == 0) {
			int len = 5;
			if (argi + 1 < argc) {
				len = atoi(argv[argi + 1]);
				if (len < 1 || len > 50) {
					return 1;
				}
			}
			return captcha_generate(len) ? 0 : 1;
		}
		if (strcmp(arg, "find") == 0) {
			if (argi + 1 < argc) {
				std::string id = argv[argi + 1];
				std::string answer = captcha_find(id.c_str());
				puts(answer.c_str());
				return 0;
			}
			return 1;
		}
		if (strcmp(arg, "verify") == 0) {
			if (argi + 2 < argc) {
				std::string id = argv[argi + 1];
				char const *in = argv[argi + 2];
				char const *p = strchr(in, ':');
				if (p) {
					std::string str(p + 1);
					int pos, len;
					if (sscanf(in, "%u,%u", &pos, &len) == 2) {
						std::string answer = captcha_find(id.c_str());
						int n = (int)answer.size();
						if (pos >= 0 && pos < n && len > 0 && pos + len <= n) {
							answer.assign(answer.c_str() + pos, len);
							if (answer == str) {
								puts("pass");
								return 0;
							}
						}
					}
				} else {
					std::string answer = captcha_find(id.c_str());
					if (answer == in) {
						puts("pass");
						return 0;
					}
				}
			}
			puts("fail");
			return 1;
		}
		if (strcmp(arg, "remove") == 0) {
			if (argi + 1 < argc) {
				char const *id = argv[argi + 1];
				return captcha_remove(id) ? 0 : 1;
			}
			return 1;
		}
		if (strcmp(arg, "test") == 0) {
			std::string text = "12345";
			response_image(text);
			return 0;
		}
	}

	std::string text;
#if 1
	{
		std::map<std::string, std::string> map;

		char const *qs = getenv("QUERY_STRING");
		if (!qs) {
			printf("Content-type: text/plain\r\n\r\n");
			return 0;
		}
		parse_query_string(qs, &map);

		std::string id;
		std::map<std::string, std::string>::const_iterator it = map.find("id");
		if (it == map.end()) {
			text = "CAPTCHA";
		} else {
			id = it->second;
			text = captcha_find(id.c_str());
		}
	}
#else
	text = "CAPTCHA";
#endif
	//

	response_image(text);

	return 0;
}

